void print_s(const char *input);
void print_i(int input);
void print_f(float input);
void print_float_arr(int rows, int cols, float arr[rows][cols]);
void print_b(size_t const size, void const *const ptr);
void print_uint32_arr_b(size_t const rows, size_t const cols, uint32_t *arr);
void print_buffer(Buffer *buffer);
void print_uint8_arr_b(size_t const rows, size_t const cols, uint8_t const *arr);